// fg_data.cxx
// Load apt.dat.gz from FlightGear, and add the airport,
// and runway points.
#include "gshhs_win.h"

#ifndef NO_ADD_FG_DATA

void Load_FG_Data( PGSINFO pgsi )
{
   PBMPINFO pbmi = &pgsi->bmpinfo;
   PGSRANGE pgsr = &pbmi->range;
   if( pgsi->fgdata.aptdat[0] ) {
      int add_all = pgsi->fgdata.add_all;
      iVerbal = (VERB9(pgsi) ? 3 : VERB5(pgsi) ? 2 : VERB3(pgsi) ? 1 : 0);
      if(VERB1(pgsi))
         sprtf( "Moment, loading %s ...\n", pgsi->fgdata.aptdat );
      int acnt = FG_Data_Load_Data_File( pgsi->fgdata.aptdat );
      if(VERB5(pgsi)) {
         sprtf( "Done. Loaded %d airports, and %s ...\n", acnt,
            (bAdd_All_Taxi ? "runways/taxiways" : "runways") );
      }
      pgsi->fgdata.apt_count = acnt;
      if(VERB5(pgsi)) {
         sprtf( "Moment, getting airports in bounds ...\n" );
      }
      FG_Data_get_apts_in_bounds( pgsr->n, pgsr->s, pgsr->e, pgsr->w,
         &pgsi->fgdata.apts, &pgsi->fgdata.rwys );
      ListCount2(&pgsi->fgdata.apts,&pgsi->fgdata.apt_cnt);
      ListCount2(&pgsi->fgdata.rwys,&pgsi->fgdata.rwy_cnt);
      if(VERB3(pgsi)) {
         sprtf( "Done. Got %d airports, %d runways%s, in bounds.\n",
            pgsi->fgdata.apt_cnt, pgsi->fgdata.rwy_cnt,
            (bAdd_All_Taxi ? "/taxiways" : "") );
         sprtf("\n");
      }
   }
   if( pgsi->fgdata.navdat[0] ) {

      if(VERB1(pgsi))
         sprtf( "Moment, loading %s ...\n", pgsi->fgdata.navdat );

      pgsi->fgdata.nav_count = Load_Nav_File( pgsi->fgdata.navdat );
      if(VERB9(pgsi)) {
         sprtf( "Done. Loaded %d navaids ...\n", pgsi->fgdata.nav_count );
         sprtf( "Moment, getting navaids in bounds ...\n" );
      }
      int bcnt = Get_Nav_in_Bounds( pgsr->n, pgsr->s, pgsr->e, pgsr->w,
         &pgsi->fgdata.navs );
      if(VERB3(pgsi)) {
         sprtf( "Done. Got %d navaids in bounds.\n", bcnt );
         sprtf("\n");
      }
   }
}

typedef struct tagNAVDETAIL {
   double latn, lonn;
   double lats, lons;
   double late, lone;
   double latw, lonw;

   double latne, lonne;
   double latse, lonse;
   double latsw, lonsw;
   double latnw, lonnw;
   double   heading;

   int ilatn, ilonn;
   int ilats, ilons;
   int ilate, ilone;
   int ilatw, ilonw;

   int ilatne, ilonne;
   int ilatse, ilonse;
   int ilatsw, ilonsw;
   int ilatnw, ilonnw;

//   int   maxx, minx, maxy, miny;
   int   xmaxx, xminx, xmaxy, xminy;

}NAVDETAIL, * PNAVDETAIL;

static NAVDETAIL navdetail;

BOOL bAddOtherPts = TRUE;

/*
from: http://www.nlreg.com/ellipse.htm
Elliptical Regression -- Fit an Ellipse to Data Points
1. The X coordinate of the center of the ellipse. 
2. The Y coordinate of the center of the ellipse. 
3. The distance from the center to the perimenter along the X axis. 
4. The distance from the center to the perimenter along the Y axis. 
5. The tilt angle of the ellipse (counter-clockwise from X axis). 

Radius = sqrt((xdim^2 * ydim^2) / ((xdim*sin(?))^2 + (ydim*cos(?))^2)) 

where Radius is the distance from the center (which in this formula is
assumed to be (0,0)), xdim is the distance from the center to the perimenter
along the X axis, ydim is the distance from the center to the perimenter
along the Y axis, and ? is the radial sweep angle counter-clockwise
from the X axis

Using the polar equation of the ellipse, we can compute the distance
from the center to a point on the perimeter for any angle. We also need
to know the distance from the center to a given (x,y) data point which 
we can calculate using Pythagorean distance formula: 

Distance = sqrt((X-Xcenter)^2 + (Y-Ycenter)^2) 

Once we know these two distances, the goal of the optimization is to
minimize the sum of the squared differences in distances (Distance - Radius). Here
is the NLREG program that performs this: 

Title "Ellipse fitted to data points";
Variables X, Y;
/ *
 *  Definite parameters to be calculated.
 *  Specifying reasonable starting values greatly increases the chances for convergence.
 * /
Parameter Xcenter = 2;      / * X coordinate of center of ellipse * /
Parameter Ycenter = 2;      / * Y coordinate of center of ellipse * /
Parameter TiltAngle = 0;    / * Rotation of ellipse in counter-clockwise direction (radians) * /
Parameter Xdim;             / * Radius of ellipse along X axis * /
Parameter Ydim;             / * Radius of ellipse along Y axis * /
/ *
 *  Work variables.
 * /
    double DataDistance, DataAngle, Deviation, Angle, r, EllipseX, EllipseY;
/ *
 *  Compute the polar rotation angle and distance of this (x,y) data point
 *  relative to the estimated center of the ellipse.
 * /
    DataAngle = rtopa(x-Xcenter, y-Ycenter);
    DataDistance = sqrt((x - Xcenter)^2 + (y - Ycenter)^2);
/ *
 *  Compute the angle for the point on the ellipse with the tilt angle.
 * /
    Angle = DataAngle - TiltAngle;
/ *
 *  Compute the radius of the ellipse (distance from center to perimeter) for
 *  this data angle. (Uses polar coordinate equation for an ellipse.)
 * /
    r = sqrt((Xdim^2 * Ydim^2) / ((Xdim*sin(Angle))^2 + (Ydim*cos(Angle))^2));
/ *
 *  Compute the difference between the distance for the data point and the ellipse.
 * /
    Deviation = DataDistance - r;
/ *
 *  Minimize the sum of squared deviations.
 * /
    Function Deviation;
/ *
 *  X, Y data values.
 * /
    Data;
[ data goes here ]
   -------------------------------- */


// p(u)=c+a*cos(u)+b*sin(u) 
// The variables a, b, c are determined when you select the endpoints
// for the first axis and the distance for the second axis. a is the 
// negative of 1/2 of the major axis length, b is the negative of 1/2
// the minor axis length, and c is the center point (2-D) of the ellipse. 
//
// Because this is actually a vector equation and the variable c is
// actually a point with X and Y values, it really should be written as: 
// p(u)=(Cx+a*cos(u))*i+(Cy+b*sin(u))*j 
// where 
// Cx is the X value of the point c 
// Cy is the Y value of the point c 
// a is -(1/2 of the major axis length) 
// b is -(1/2 of the minor axis length) 
// i and j represent unit vectors in the X and Y directions 
// OR
// (x^2 x b^2) + (y^2 x a^2) = a^2 x b^2
// where a and b are the axes - such that
// (x^2 / a^2) + (y^2 / b^2) = 1 or
// (x^2 / b^2) + (y^2 / a^2) = 1 or
// Eccentricity =
// E = sqrt( a^2 - b^2 ) / a;
// x^2 / a^2 = (a^2 x b^2) - (y^2 / b^2)
// x = sqrt( a^2 * ((a^2 x b^2) - (y^2 / b^2)) );

int ADD2BMP( PGSINFO pgsi, int y, int x, BYTE b, BOOL ovr )
{
   int   iret = 0;
   if( Pt_In_BMP_Boundary( pgsi, y, x ) ) {
      if( Add_Pt_2_BMP( pgsi, y, x, b, ovr ) )
         iret = 1;
   }
   return iret;
}
int Add_ILS_Item_2_BMP( PGSINFO pgsi, PNAVITEM pni )
{
   int   added2 = 0;
   double lat = pni->lat;
   double lon = pni->lon;
   char * pident = pni->ident;
   int idlen = (int)strlen(pident);
   int   y, x;

   Get_BMP_iLatLon( pgsi, lat, lon, &y, &x );
   if( ADD2BMP( pgsi, y, x, BMC_FUCHIA, FALSE ) )
      added2++;

   return added2;
}

int Add_OTHER_Item_2_BMP( PGSINFO pgsi, PNAVITEM pni )
{
   int   added2 = 0;
   double lat = pni->lat;
   double lon = pni->lon;
   char * pident = pni->ident;
   int idlen = (int)strlen(pident);
   int   y, x;

   Get_BMP_iLatLon( pgsi, lat, lon, &y, &x );
   if( ADD2BMP( pgsi, y, x, BMC_AQUA, FALSE ) )
      added2++;

   return added2;
}


int Add_NAV_Item_2_BMP( PGSINFO pgsi, PNAVITEM pni )
{
   PNAVDETAIL pnd = &navdetail;
   int   added2 = 0;
   double mets;
   double lat = pni->lat;
   double lon = pni->lon;
   double range = (double)pni->range;  // in nm
   RECT  rc, rc2;
   char * pident = pni->ident;
   int idlen = (int)strlen(pident);
   int   addc = pgsi->fgdata.add_circle;
   int   addi = pgsi->fgdata.add_ident;

   mets = range * SG_NM_TO_METER;
   geo_direct_wgs_84 ( lat, lon, 0.0, mets,
      &pnd->latn, &pnd->lonn, &pnd->heading );
   geo_direct_wgs_84 ( lat, lon, 45.0, mets,
      &pnd->latne, &pnd->lonne, &pnd->heading );
   geo_direct_wgs_84 ( lat, lon, 90.0, mets,
      &pnd->late, &pnd->lone, &pnd->heading );
   geo_direct_wgs_84 ( lat, lon, 135.0, mets,
      &pnd->latse, &pnd->lonse, &pnd->heading );
   geo_direct_wgs_84 ( lat, lon, 180.0, mets,
      &pnd->lats, &pnd->lons, &pnd->heading );
   geo_direct_wgs_84 ( lat, lon, 225.0, mets,
      &pnd->latsw, &pnd->lonsw, &pnd->heading );
   geo_direct_wgs_84 ( lat, lon, 270.0, mets,
      &pnd->latw, &pnd->lonw, &pnd->heading );
   geo_direct_wgs_84 ( lat, lon, 315.0, mets,
      &pnd->latnw, &pnd->lonnw, &pnd->heading );

   Get_BMP_iLatLon( pgsi, pnd->latn, pnd->lonn, &pnd->ilatn, &pnd->ilonn );
   Get_BMP_iLatLon( pgsi, pnd->late, pnd->lone, &pnd->ilate, &pnd->ilone );
   Get_BMP_iLatLon( pgsi, pnd->lats, pnd->lons, &pnd->ilats, &pnd->ilons );
   Get_BMP_iLatLon( pgsi, pnd->latw, pnd->lonw, &pnd->ilatw, &pnd->ilonw );
   Get_BMP_iLatLon( pgsi, pnd->latne, pnd->lonne, &pnd->ilatne, &pnd->ilonne );
   Get_BMP_iLatLon( pgsi, pnd->latse, pnd->lonse, &pnd->ilatse, &pnd->ilonse );
   Get_BMP_iLatLon( pgsi, pnd->latsw, pnd->lonsw, &pnd->ilatsw, &pnd->ilonsw );
   Get_BMP_iLatLon( pgsi, pnd->latnw, pnd->lonnw, &pnd->ilatnw, &pnd->ilonnw );

   if( addc ) {
      if( ADD2BMP( pgsi, pnd->ilatn, pnd->ilonn, BMC_GOLD, FALSE ) )
         added2++;
      if( ADD2BMP( pgsi, pnd->ilatne, pnd->ilonne, BMC_GOLD, FALSE ) )
         added2++;
      if( ADD2BMP( pgsi, pnd->ilate, pnd->ilone, BMC_GOLD, FALSE ) )
         added2++;
      if( ADD2BMP( pgsi, pnd->ilatse, pnd->ilonse, BMC_GOLD, FALSE ) )
         added2++;
      if( ADD2BMP( pgsi, pnd->ilats, pnd->ilons, BMC_GOLD, FALSE ) )
         added2++;
      if( ADD2BMP( pgsi, pnd->ilatsw, pnd->ilonsw, BMC_GOLD, FALSE ) )
         added2++;
      if( ADD2BMP( pgsi, pnd->ilatw, pnd->ilonw, BMC_GOLD, FALSE ) )
         added2++;
      if( ADD2BMP( pgsi, pnd->ilatnw, pnd->ilonnw, BMC_GOLD, FALSE ) )
         added2++;
   }

   rc.left = pnd->ilonw;   // west point
   rc.top  = pnd->ilatn;   // north point
   rc.right = pnd->ilone;  // east point
   rc.bottom = pnd->ilats; // south point
   rc2 = rc;
   NormalizeRect(&rc2);
   if( VERB9(pgsi) )
      sprtf( "Rect1: %s, 2 %s\n", Rect2StgFull(&rc), Rect2StgFull(&rc2) );

   if(( rc2.left < rc2.right )&&
      ( rc2.top < rc2.bottom )) {
      int x, y;
      int cx, cy;
      int hy, hx, iy, ix;
      double dx, dy, ecc;
      double a, b, a2, b2;
      PPOINT ppt;
      int   icnt = 0;
      int   repeated = 0;
      int   i, j;
      int ptcnt;

      cx = rc2.right - rc2.left;
      cy = rc2.bottom - rc2.top;
      if( cx < cy ) {
         x = cy;
         cy = cx;
         cx = x;
      }
      dx = (double)cx;
      dy = (double)cy;
      a = dx / 2.0;
      b = dy / 2.0;
      a2 = a * a;
      b2 = b * b;
      ecc = sqrt( (dx * dx) - (dy * dy) ) / dx;
      if( VERB9(pgsi) )
         sprtf( "Axes: a = %d, b = %d, Eccentricity = %3.2f\n", cx, cy, ecc );
      hy = ((cy + 1) / 2);
      hx = ((cx + 1) / 2);
      ppt = (PPOINT)MALLOC( sizeof(POINT) * (hx+1) * (hy+1) );
      CHKMEM(ppt, (hx+1)*(hy+1)*sizeof(POINT) );
      rc.bottom = rc2.top + hy;
      rc.right = rc2.left + hx;
      for( y = rc2.top; y < rc.bottom; y++ ) {
         dy = (double) (y - rc2.top);
         // (x2 / a2) + (y2 / b2) = 1
         // (x2 / a2) = 1 - (y2 / b2)
         // x = sqrt( a2( 1 - (y2 / b2)))
         dx = sqrt( a2 * ( 1.0 - ((dy * dy) / b2)) );
         ix = (int)(dx + 0.5);
         iy = (int)dy;
         //sprtf( "Point y = %d, then x = %d ...(%d)\n", iy, ix, y );
         ppt[icnt].x = ix;
         ppt[icnt].y = iy;
         icnt++;
      }
      if( VERB9(pgsi) )
         sprtf( "Last Point y = %d, then x = %d ...(%d)\n", iy, ix, y-1 );
      for( x = rc2.left; x < rc.right; x++)
      {
         dx = (double)( x - rc2.left );
         // (x2 / a2) + (y2 / b2) = 1
         // y2 / b2 = 1 - (x2 / a2)
         // y2 = b2 * ( 1 - (x2 / a2) )
         dy = sqrt( b2 * ( 1.0 - (( dx * dx ) / a2) ) );
         iy = (int)(dy + 0.5);
         ix = (int)dx;
         //sprtf( "Point x = %d, then y = %d ...(%d)\n", ix, iy, x );
         for( j = 0; j < icnt; j++ )
         {
            if(( ppt[j].x == ix )&&
               ( ppt[j].y == iy ))
               break;

         }
         if( j < icnt ) {
            repeated++;
         } else {
            ppt[icnt].x = ix;
            ppt[icnt].y = iy;
            icnt++;
         }
      }
      if( VERB9(pgsi) )
         sprtf( "Last Point x = %d, then y = %d ...(%d)\n", ix, iy, x-1 );

      pnd->xmaxx = _MIN_INT32;
      pnd->xmaxy = _MIN_INT32;
      pnd->xminx = _MAX_INT32;
      pnd->xminy = _MAX_INT32;
      // adjust to center
      rc2.left += hx;
      rc2.top  += hy;
      for(i = 0; i < icnt; i++)
      {
         x = ppt[i].x;
         y = ppt[i].y;
         x += rc2.left;
         y += rc2.top;
         if( addc ) {
            if( ADD2BMP( pgsi, y, x, BMC_GOLD, FALSE ) ) {
               added2++;
            }
         }
         if( x > pnd->xmaxx )
            pnd->xmaxx = x;
         if( x < pnd->xminx )
            pnd->xminx = x;
         if( y > pnd->xmaxy )
            pnd->xmaxy = y;
         if( y < pnd->xminy )
            pnd->xminy = y;

         x = -ppt[i].x;
         y = ppt[i].y;
         x += rc2.left;
         y += rc2.top;
         if( addc ) {
            if( ADD2BMP( pgsi, y, x, BMC_GOLD, FALSE ) ) {
               added2++;
            }
         }
         if( x > pnd->xmaxx )
            pnd->xmaxx = x;
         if( x < pnd->xminx )
            pnd->xminx = x;
         if( y > pnd->xmaxy )
            pnd->xmaxy = y;
         if( y < pnd->xminy )
            pnd->xminy = y;

         x = -ppt[i].x;
         y = -ppt[i].y;
         x += rc2.left;
         y += rc2.top;
         if( addc ) {
            if( ADD2BMP( pgsi, y, x, BMC_GOLD, FALSE ) ) {
               added2++;
            }
         }
         if( x > pnd->xmaxx )
            pnd->xmaxx = x;
         if( x < pnd->xminx )
            pnd->xminx = x;
         if( y > pnd->xmaxy )
            pnd->xmaxy = y;
         if( y < pnd->xminy )
            pnd->xminy = y;

         x = ppt[i].x;
         y = -ppt[i].y;
         x += rc2.left;
         y += rc2.top;
         if( addc ) {
            if( ADD2BMP( pgsi, y, x, BMC_GOLD, FALSE ) ) {
               added2++;
            }
         }
         if( x > pnd->xmaxx )
            pnd->xmaxx = x;
         if( x < pnd->xminx )
            pnd->xminx = x;
         if( y > pnd->xmaxy )
            pnd->xmaxy = y;
         if( y < pnd->xminy )
            pnd->xminy = y;
      }

      if( VERB9(pgsi) ) {
         sprtf( "xmaxx %d, xminx %d, xmaxy %d, xminy %d\n",
            pnd->xmaxx, pnd->xminx, pnd->xmaxy, pnd->xminy );
      }
      y = (pnd->xminy + pnd->xmaxy) / 2;
      if( addi ) {
         x = pnd->xminx;
         Add_Text_to_BMP( pgsi, x, y, pident, idlen, BMC_YELLOW );
         x = pnd->xmaxx;
         Add_Text_to_BMP( pgsi, x, y, pident, idlen, BMC_YELLOW );
      }
      if( addc ) {
         ptcnt = Add_Line_to_BMP( pgsi, pnd->xminx, y, pnd->xmaxx, y, BMC_YELLOW, FALSE );
         if( VERB9(pgsi) ) {
            sprtf( "Added horizontal line at y=%d, from x %d-%d, %d points BMC_YELLOW"MEOR,
               y, pnd->xminx, pnd->xmaxx, ptcnt );
         }
      }

      x = (pnd->xminx + pnd->xmaxx) / 2;
      if( addi ) {
         y = pnd->xmaxy;
         Add_Text_to_BMP( pgsi, x, y, pident, idlen, BMC_YELLOW );
         y = pnd->xminy;
         Add_Text_to_BMP( pgsi, x, y, pident, idlen, BMC_YELLOW );
      }
      if( addc ) {
         ptcnt = Add_Line_to_BMP( pgsi, x, pnd->xminy, x, pnd->xmaxy, BMC_YELLOW, FALSE );
         if( VERB9(pgsi) ) {
            sprtf( "Added vertical line at x=%d, from y %d-%d, %d points BMC_YELLOW"MEOR,
               y, pnd->xminy, pnd->xmaxy, ptcnt );
         }
      }
   }

   return added2;
}



void Add_FG_Navs( PGSINFO pgsi )
{
   PLE ph = &pgsi->fgdata.navs;
   PLE pn;
   PNAVDETAIL pnd = &navdetail;
   PNAVITEM pni;
   int   added = 0;
   int   added2 = 0;
   double lat, lon;
   int   ilat, ilon;
   int   len;
   int   addi = pgsi->fgdata.add_ident;
   int   addc = pgsi->fgdata.add_circle;

   if( VERB2(pgsi) ) {
      sprtf( "Adding navaids to BMP ... VOR/NDB=%s, ID=%s\n",
         (addc ? "On" : "Off"),
         (addi ? "On" : "Off") );
   }
   Traverse_List(ph,pn)
   {
      pni = (PNAVITEM)pn;
      lat = pni->lat;
      lon = pni->lon;
      Get_BMP_iLatLon( pgsi, lat, lon, &ilat, &ilon );
      switch(pni->fgtype)
      {
      case FG_NAV_VOR:
      case FG_NAV_NDB:
         // paint a circle, per the range
         if( VERB5(pgsi) )
            sprtf( "Item: VOR/NDB id[%s] name[%s] ...\n", pni->ident, pni->name );
         if(addi) {
            len = (int)strlen(pni->ident);
            Add_Text_to_BMP( pgsi, ilon, ilat, pni->ident, len, BMC_MAROON );
         }
         break;
      case FG_NAV_ILS:
         if( VERB5(pgsi) )
            sprtf( "Item: ILS %s ...\n", pni->name );
         break;
      default:
         if( VERB5(pgsi) )
            sprtf( "Item: NAV Other %s ...\n", pni->name );
         break;
      }

      if( Add_Pt_2_BMP( pgsi, ilat, ilon, BMC_GOLD, FALSE ) )
         added++;


      if( bAddOtherPts ) {
         //int fgtype = Get_Nav_Type( pni->type );
         switch(pni->fgtype)
         {
         case FG_NAV_VOR:
         case FG_NAV_NDB:
            // paint a circle, per the range
            added2 += Add_NAV_Item_2_BMP( pgsi, pni );
            break;
         case FG_NAV_ILS:
            // paint radial lines, per direction
            added2 += Add_ILS_Item_2_BMP( pgsi, pni );
            break;
         default:
            added2 += Add_OTHER_Item_2_BMP( pgsi, pni );
            break;
         }
      }
   }
   if( VERB3(pgsi) )
      sprtf( "Added %d navaids to BMP, with %d other points ...\n", added, added2 );
}

void Paint_FG_Data( PGSINFO pgsi )
{
   PLE ph1 = &pgsi->fgdata.apts;
   PLE ph2 = &pgsi->fgdata.rwys;
   PLE ph3 = &pgsi->fgdata.navs;
   if( !IsListEmpty(ph1) ) {
      Work_With_List( pgsi );
   }
   if( !IsListEmpty(ph3) ) {
      Add_FG_Navs( pgsi );
   }
}

void Paint_FG_World( PGSINFO pgsi )
{
   PLE phrwys = Get_Runways_List();
   PLE phapts = Get_Airports_List();
   PLE phnavs = Get_Nav_List();
   PLE   pn;
   int   rcnt = 0;
   int   acnt = 0;
   int   ncnt = 0;
   if(VERB3(pgsi))
      sprtf( "Adding airports and runways to World ...\n" );
   Traverse_List( phapts, pn )
   {
      PFGAPT papt = (PFGAPT)pn;
      acnt++;
      Add_2_World( pgsi, papt->lon, papt->lat, BMC_MAROON, FALSE );
   }
   Traverse_List( phrwys, pn )
   {
      PFGRWY prwy = (PFGRWY)pn;
      rcnt++;
      Add_2_World( pgsi, prwy->lon, prwy->lat, BMC_OLIVE, FALSE );
   }
   if(VERB3(pgsi))
      sprtf( "Done %d airports, and %d runways to World ...\n", acnt, rcnt );
   Traverse_List( phnavs, pn )
   {
      PNAVITEM pni = (PNAVITEM)pn;
      ncnt++;
      Add_2_World( pgsi, pni->lon, pni->lat, BMC_GOLD, FALSE );
   }
   if(VERB3(pgsi))
      sprtf( "Done %d navaids to World ...\n", ncnt );
}

#endif // #ifndef NO_ADD_FG_DATA
// eof - fg_data.cxx
