// nav_dat.h
#ifndef _nav_dat_H_
#define _nav_dat_H_
#ifdef __cplusplus
extern "C" {
#endif

#define FG_NAV_DEFAULT_RANGE 50 // nm
#define FG_LOC_DEFAULT_RANGE 18 // nm
#define FG_DME_DEFAULT_RANGE 50 // nm
#define FG_NAV_MAX_RANGE 300    // nm

// Shield the rest of FG from possibly changing details of Robins navaid type numbering system.
// Currently only the GPS code uses this - extra types (LOC, GS etc) may need to be added
// should other FG code choose to use this. 
typedef enum  {
    FG_NAV_VOR,
    FG_NAV_NDB,
    FG_NAV_ILS,
    FG_NAV_ANY
}FG_NAV_TYPES;

#define  MX_NAV_ID   8
#define  MX_NAV_NAME 48

typedef struct tagNAVITEM {
   LE    link;
   int type;   // number of types
   FG_NAV_TYPES fgtype; // specific FG enumeration
   double lat, lon;  // location
   int   elev_ft;
   int   freq;
   int   range;   // in nm
   double multi;
   char ident[MX_NAV_ID+1];
   char aptid[MX_NAV_ID+1];
   char name[MX_NAV_NAME+1];
}NAVITEM, * PNAVITEM;

extern int Load_Nav_File( char * pfile );
extern PLE Get_Nav_List( void );
extern void Del_Nav_List( void );
extern int Get_Nav_in_Bounds( double n, double s, double e, double w, PLE ph );
extern FG_NAV_TYPES Get_Nav_Type( int type );

#ifdef __cplusplus
}
#endif
#endif _nav_dat_H_
// eof - nav_dat.h
